import pytest

from xpoz import PaginatedResult
from xpoz.types.instagram import InstagramPost, InstagramUser, InstagramComment
from xpoz.types.common import PaginationInfo


@pytest.fixture(scope="module")
def instagram_posts_result(client):
    return client.instagram.get_posts_by_user(
        "instagram", fields=["id", "caption", "like_count"]
    )


@pytest.fixture(scope="module")
def instagram_post_id():
    return "3650461835687763021_8763092944"


class TestInstagramUsers:
    def test_get_user(self, client):
        user = client.instagram.get_user("instagram")
        assert isinstance(user, InstagramUser)
        assert user.username == "instagram"

    def test_get_user_with_fields(self, client):
        user = client.instagram.get_user(
            "instagram", fields=["id", "username", "follower_count"]
        )
        assert isinstance(user, InstagramUser)
        assert user.id is not None
        assert user.username is not None
        assert user.follower_count is not None

    def test_search_users(self, client):
        users = client.instagram.search_users("nike")
        assert isinstance(users, list)
        assert len(users) > 0
        for u in users:
            assert isinstance(u, InstagramUser)

    def test_get_user_connections(self, client):
        result = client.instagram.get_user_connections("instagram", "followers")
        assert isinstance(result, PaginatedResult)
        assert isinstance(result.pagination, PaginationInfo)
        assert len(result.data) > 0
        for u in result.data:
            assert isinstance(u, InstagramUser)

    def test_get_users_by_keywords(self, client):
        result = client.instagram.get_users_by_keywords("fashion")
        assert isinstance(result, PaginatedResult)
        assert len(result.data) > 0
        for u in result.data:
            assert isinstance(u, InstagramUser)


class TestInstagramPosts:
    def test_get_posts_by_user(self, instagram_posts_result):
        result = instagram_posts_result
        assert isinstance(result, PaginatedResult)
        assert len(result.data) > 0
        for post in result.data:
            assert isinstance(post, InstagramPost)

    def test_search_posts(self, client):
        result = client.instagram.search_posts(
            "travel", fields=["id", "caption", "like_count"]
        )
        assert isinstance(result, PaginatedResult)
        assert isinstance(result.pagination, PaginationInfo)
        assert result.pagination.total_rows > 0

    def test_get_posts_by_ids(self, client, instagram_post_id):
        posts = client.instagram.get_posts_by_ids([instagram_post_id])
        assert isinstance(posts, list)
        assert len(posts) == 1
        assert isinstance(posts[0], InstagramPost)

    def test_get_comments(self, client, instagram_post_id):
        result = client.instagram.get_comments(instagram_post_id)
        assert isinstance(result, PaginatedResult)

    def test_get_post_interacting_users(self, client, instagram_post_id):
        result = client.instagram.get_post_interacting_users(
            instagram_post_id, "commenters"
        )
        assert isinstance(result, PaginatedResult)



from __future__ import annotations

from typing import Any

from xpoz.namespaces._base import (
    BaseNamespace,
    AsyncBaseNamespace,
    _parse_item,
    _parse_items,
    _extract_pagination,
    _extract_results,
)
from xpoz._pagination import PaginatedResult, AsyncPaginatedResult
from xpoz._mcp._polling import wait_for_result_sync, wait_for_result
from xpoz.types.reddit import (
    RedditPost,
    RedditUser,
    RedditComment,
    RedditSubreddit,
    RedditPostWithComments,
    SubredditWithPosts,
)
from xpoz.types.common import PaginationInfo
from xpoz._config import _tools


class RedditNamespace(BaseNamespace):
    def search_posts(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        sort: str | None = None,
        time: str | None = None,
        subreddit: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[RedditPost]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            sort=sort,
            time=time,
            subreddit=subreddit,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.SEARCH_REDDIT_POSTS, args)
        return self._build_paginated_result(
            result, RedditPost, _tools.SEARCH_REDDIT_POSTS, args
        )

    def get_post_with_comments(
        self,
        post_id: str,
        *,
        post_fields: list[str] | None = None,
        comment_fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> RedditPostWithComments:
        args = self._build_args(
            postId=post_id,
            postFields=self._convert_fields(post_fields),
            commentFields=self._convert_fields(comment_fields),
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_REDDIT_POST_WITH_COMMENTS, args)
        return self._parse_post_with_comments(result)

    def search_comments(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        subreddit: str | None = None,
    ) -> PaginatedResult[RedditComment]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            subreddit=subreddit,
        )
        result = self._call_and_maybe_poll(_tools.SEARCH_REDDIT_COMMENTS, args)
        return self._build_paginated_result(
            result, RedditComment, _tools.SEARCH_REDDIT_COMMENTS, args
        )

    def get_user(
        self,
        username: str,
        *,
        fields: list[str] | None = None,
    ) -> RedditUser:
        args = self._build_args(
            username=username,
            fields=self._convert_fields(fields),
        )
        result = self._call_and_maybe_poll(_tools.GET_REDDIT_USER, args)
        results = result.get("results", [])
        if isinstance(results, list) and len(results) > 0:
            return _parse_item(RedditUser, results[0])
        return _parse_item(RedditUser, result)

    def search_users(
        self,
        name: str,
        *,
        limit: int | None = None,
        fields: list[str] | None = None,
    ) -> list[RedditUser]:
        args = self._build_args(
            name=name,
            limit=limit,
            fields=self._convert_fields(fields),
        )
        result = self._call_and_maybe_poll(_tools.SEARCH_REDDIT_USERS, args)
        return _parse_items(RedditUser, result.get("results", []))

    def get_users_by_keywords(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        subreddit: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[RedditUser]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            subreddit=subreddit,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_REDDIT_USERS_BY_KEYWORDS, args)
        return self._build_paginated_result(
            result, RedditUser, _tools.GET_REDDIT_USERS_BY_KEYWORDS, args
        )

    def search_subreddits(
        self,
        query: str,
        *,
        limit: int | None = None,
        fields: list[str] | None = None,
    ) -> list[RedditSubreddit]:
        args = self._build_args(
            query=query,
            limit=limit,
            fields=self._convert_fields(fields),
        )
        result = self._call_and_maybe_poll(_tools.SEARCH_REDDIT_SUBREDDITS, args)
        return _parse_items(RedditSubreddit, result.get("results", []))

    def get_subreddit_with_posts(
        self,
        subreddit_name: str,
        *,
        subreddit_fields: list[str] | None = None,
        post_fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> SubredditWithPosts:
        args = self._build_args(
            subredditName=subreddit_name,
            subredditFields=self._convert_fields(subreddit_fields),
            postFields=self._convert_fields(post_fields),
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_REDDIT_SUBREDDIT_WITH_POSTS, args)
        return self._parse_subreddit_with_posts(result)

    def get_subreddits_by_keywords(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[RedditSubreddit]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_REDDIT_SUBREDDITS_BY_KEYWORDS, args)
        return self._build_paginated_result(
            result, RedditSubreddit, _tools.GET_REDDIT_SUBREDDITS_BY_KEYWORDS, args
        )

    def _parse_post_with_comments(self, raw: dict[str, Any]) -> RedditPostWithComments:
        results = raw.get("results", {})
        if isinstance(results, list):
            results = results[0] if results else {}

        post_data = results.get("post", results)
        comments_data = results.get("comments") or []
        pagination = _extract_pagination(raw)

        return RedditPostWithComments(
            post=_parse_item(RedditPost, post_data),
            comments=_parse_items(RedditComment, comments_data),
            comments_pagination=pagination if pagination.total_pages > 0 else None,
            comments_table_name=pagination.table_name,
        )

    def _parse_subreddit_with_posts(self, raw: dict[str, Any]) -> SubredditWithPosts:
        results = raw.get("results", {})
        if isinstance(results, list):
            results = results[0] if results else {}

        subreddit_data = results.get("subreddit", results)
        posts_data = results.get("posts") or []
        pagination = _extract_pagination(raw)

        return SubredditWithPosts(
            subreddit=_parse_item(RedditSubreddit, subreddit_data),
            posts=_parse_items(RedditPost, posts_data),
            posts_pagination=pagination if pagination.total_pages > 0 else None,
            posts_table_name=pagination.table_name,
        )


class AsyncRedditNamespace(AsyncBaseNamespace):
    async def search_posts(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        sort: str | None = None,
        time: str | None = None,
        subreddit: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[RedditPost]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            sort=sort,
            time=time,
            subreddit=subreddit,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.SEARCH_REDDIT_POSTS, args)
        return await self._build_paginated_result(
            result, RedditPost, _tools.SEARCH_REDDIT_POSTS, args
        )

    async def get_post_with_comments(
        self,
        post_id: str,
        *,
        post_fields: list[str] | None = None,
        comment_fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> RedditPostWithComments:
        args = self._build_args(
            postId=post_id,
            postFields=self._convert_fields(post_fields),
            commentFields=self._convert_fields(comment_fields),
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_REDDIT_POST_WITH_COMMENTS, args)
        return self._parse_post_with_comments(result)

    async def search_comments(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        subreddit: str | None = None,
    ) -> AsyncPaginatedResult[RedditComment]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            subreddit=subreddit,
        )
        result = await self._call_and_maybe_poll(_tools.SEARCH_REDDIT_COMMENTS, args)
        return await self._build_paginated_result(
            result, RedditComment, _tools.SEARCH_REDDIT_COMMENTS, args
        )

    async def get_user(
        self,
        username: str,
        *,
        fields: list[str] | None = None,
    ) -> RedditUser:
        args = self._build_args(
            username=username,
            fields=self._convert_fields(fields),
        )
        result = await self._call_and_maybe_poll(_tools.GET_REDDIT_USER, args)
        results = result.get("results", [])
        if isinstance(results, list) and len(results) > 0:
            return _parse_item(RedditUser, results[0])
        return _parse_item(RedditUser, result)

    async def search_users(
        self,
        name: str,
        *,
        limit: int | None = None,
        fields: list[str] | None = None,
    ) -> list[RedditUser]:
        args = self._build_args(
            name=name,
            limit=limit,
            fields=self._convert_fields(fields),
        )
        result = await self._call_and_maybe_poll(_tools.SEARCH_REDDIT_USERS, args)
        return _parse_items(RedditUser, result.get("results", []))

    async def get_users_by_keywords(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        subreddit: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[RedditUser]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            subreddit=subreddit,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_REDDIT_USERS_BY_KEYWORDS, args)
        return await self._build_paginated_result(
            result, RedditUser, _tools.GET_REDDIT_USERS_BY_KEYWORDS, args
        )

    async def search_subreddits(
        self,
        query: str,
        *,
        limit: int | None = None,
        fields: list[str] | None = None,
    ) -> list[RedditSubreddit]:
        args = self._build_args(
            query=query,
            limit=limit,
            fields=self._convert_fields(fields),
        )
        result = await self._call_and_maybe_poll(_tools.SEARCH_REDDIT_SUBREDDITS, args)
        return _parse_items(RedditSubreddit, result.get("results", []))

    async def get_subreddit_with_posts(
        self,
        subreddit_name: str,
        *,
        subreddit_fields: list[str] | None = None,
        post_fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> SubredditWithPosts:
        args = self._build_args(
            subredditName=subreddit_name,
            subredditFields=self._convert_fields(subreddit_fields),
            postFields=self._convert_fields(post_fields),
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_REDDIT_SUBREDDIT_WITH_POSTS, args)
        return self._parse_subreddit_with_posts(result)

    async def get_subreddits_by_keywords(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[RedditSubreddit]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_REDDIT_SUBREDDITS_BY_KEYWORDS, args)
        return await self._build_paginated_result(
            result, RedditSubreddit, _tools.GET_REDDIT_SUBREDDITS_BY_KEYWORDS, args
        )

    def _parse_post_with_comments(self, raw: dict[str, Any]) -> RedditPostWithComments:
        results = raw.get("results", {})
        if isinstance(results, list):
            results = results[0] if results else {}

        post_data = results.get("post", results)
        comments_data = results.get("comments") or []
        pagination = _extract_pagination(raw)

        return RedditPostWithComments(
            post=_parse_item(RedditPost, post_data),
            comments=_parse_items(RedditComment, comments_data),
            comments_pagination=pagination if pagination.total_pages > 0 else None,
            comments_table_name=pagination.table_name,
        )

    def _parse_subreddit_with_posts(self, raw: dict[str, Any]) -> SubredditWithPosts:
        results = raw.get("results", {})
        if isinstance(results, list):
            results = results[0] if results else {}

        subreddit_data = results.get("subreddit", results)
        posts_data = results.get("posts") or []
        pagination = _extract_pagination(raw)

        return SubredditWithPosts(
            subreddit=_parse_item(RedditSubreddit, subreddit_data),
            posts=_parse_items(RedditPost, posts_data),
            posts_pagination=pagination if pagination.total_pages > 0 else None,
            posts_table_name=pagination.table_name,
        )



import os

import pytest

from xpoz import XpozClient


@pytest.fixture(scope="module")
def client():
    api_key = os.environ.get("XPOZ_API_KEY")
    server_url = os.environ.get("XPOZ_SERVER_URL")
    if not api_key:
        pytest.skip("XPOZ_API_KEY not set")
    c = XpozClient(api_key=api_key, server_url=server_url, timeout=600)
    yield c
    c.close()



from __future__ import annotations

from typing import Any

from xpoz.namespaces._base import BaseNamespace, AsyncBaseNamespace, _parse_item, _parse_items
from xpoz._pagination import PaginatedResult, AsyncPaginatedResult
from xpoz.types.twitter import TwitterPost, TwitterUser
from xpoz._config import _tools


class TwitterNamespace(BaseNamespace):
    def get_posts_by_ids(
        self,
        post_ids: list[str],
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> list[TwitterPost]:
        args = self._build_args(
            postIds=post_ids,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_POSTS_BY_IDS, args)
        return _parse_items(TwitterPost, result.get("results", []))

    def get_posts_by_author(
        self,
        identifier: str,
        identifier_type: str = "username",
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[TwitterPost]:
        args = self._build_args(
            identifier=identifier,
            identifierType=identifier_type,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_POSTS_BY_AUTHOR, args)
        return self._build_paginated_result(result, TwitterPost, _tools.GET_TWITTER_POSTS_BY_AUTHOR, args)

    def search_posts(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        author_username: str | None = None,
        author_id: str | None = None,
        language: str | None = None,
        force_latest: bool | None = None,
        response_type: str | None = None,
    ) -> PaginatedResult[TwitterPost]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            authorUsername=author_username,
            authorId=author_id,
            language=language,
            forceLatest=force_latest,
            responseType=response_type,
        )
        if response_type == "csv":
            raw = self._call_tool(_tools.SEARCH_TWITTER_POSTS, args)
            export_op_id = raw.get("operationId") or raw.get("dataDumpExportOperationId")
            csv_raw: dict[str, Any] = {
                "results": [],
                "dataDumpExportOperationId": export_op_id,
            }
            return self._build_paginated_result(csv_raw, TwitterPost, _tools.SEARCH_TWITTER_POSTS, args)
        result = self._call_and_maybe_poll(_tools.SEARCH_TWITTER_POSTS, args)
        return self._build_paginated_result(result, TwitterPost, _tools.SEARCH_TWITTER_POSTS, args)

    def get_retweets(
        self,
        post_id: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
    ) -> PaginatedResult[TwitterPost]:
        args = self._build_args(
            postId=post_id,
            fields=self._convert_fields(fields),
            startDate=start_date,
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_RETWEETS, args)
        return self._build_paginated_result(result, TwitterPost, _tools.GET_TWITTER_RETWEETS, args)

    def get_quotes(
        self,
        post_id: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[TwitterPost]:
        args = self._build_args(
            postId=post_id,
            fields=self._convert_fields(fields),
            startDate=start_date,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_QUOTES, args)
        return self._build_paginated_result(result, TwitterPost, _tools.GET_TWITTER_QUOTES, args)

    def get_comments(
        self,
        post_id: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[TwitterPost]:
        args = self._build_args(
            postId=post_id,
            fields=self._convert_fields(fields),
            startDate=start_date,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_COMMENTS, args)
        return self._build_paginated_result(result, TwitterPost, _tools.GET_TWITTER_COMMENTS, args)

    def get_post_interacting_users(
        self,
        post_id: str,
        interaction_type: str,
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[TwitterUser]:
        args = self._build_args(
            postId=post_id,
            interactionType=interaction_type,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_POST_INTERACTING_USERS, args)
        return self._build_paginated_result(
            result, TwitterUser, _tools.GET_TWITTER_POST_INTERACTING_USERS, args
        )

    def count_posts(
        self,
        phrase: str,
        *,
        start_date: str | None = None,
        end_date: str | None = None,
    ) -> int:
        args = self._build_args(
            phrase=phrase,
            startDate=start_date,
            endDate=end_date,
        )
        result = self._call_and_maybe_poll(_tools.COUNT_TWEETS, args)
        count = result.get("results", result.get("count", 0))
        if isinstance(count, list):
            if len(count) == 0:
                return 0
            first = count[0]
            if isinstance(first, dict):
                return int(next(iter(first.values())))
            return int(first)
        return int(count)

    def get_user(
        self,
        identifier: str,
        identifier_type: str = "username",
        *,
        fields: list[str] | None = None,
    ) -> TwitterUser:
        args = self._build_args(
            identifier=identifier,
            identifierType=identifier_type,
            fields=self._convert_fields(fields),
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_USER, args)
        results = result.get("results", [])
        if isinstance(results, list) and len(results) > 0:
            return _parse_item(TwitterUser, results[0])
        return _parse_item(TwitterUser, result)

    def search_users(
        self,
        name: str,
        *,
        limit: int | None = None,
        fields: list[str] | None = None,
    ) -> list[TwitterUser]:
        args = self._build_args(
            name=name,
            limit=limit,
            fields=self._convert_fields(fields),
        )
        result = self._call_and_maybe_poll(_tools.SEARCH_TWITTER_USERS, args)
        return _parse_items(TwitterUser, result.get("results", []))

    def get_user_connections(
        self,
        username: str,
        connection_type: str,
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[TwitterUser]:
        args = self._build_args(
            username=username,
            connectionType=connection_type,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_USER_CONNECTIONS, args)
        return self._build_paginated_result(
            result, TwitterUser, _tools.GET_TWITTER_USER_CONNECTIONS, args
        )

    def get_users_by_keywords(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        language: str | None = None,
        force_latest: bool | None = None,
    ) -> PaginatedResult[TwitterUser]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            language=language,
            forceLatest=force_latest,
        )
        result = self._call_and_maybe_poll(_tools.GET_TWITTER_USERS_BY_KEYWORDS, args)
        return self._build_paginated_result(
            result, TwitterUser, _tools.GET_TWITTER_USERS_BY_KEYWORDS, args
        )


class AsyncTwitterNamespace(AsyncBaseNamespace):
    async def get_posts_by_ids(
        self,
        post_ids: list[str],
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> list[TwitterPost]:
        args = self._build_args(
            postIds=post_ids,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_POSTS_BY_IDS, args)
        return _parse_items(TwitterPost, result.get("results", []))

    async def get_posts_by_author(
        self,
        identifier: str,
        identifier_type: str = "username",
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[TwitterPost]:
        args = self._build_args(
            identifier=identifier,
            identifierType=identifier_type,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_POSTS_BY_AUTHOR, args)
        return await self._build_paginated_result(
            result, TwitterPost, _tools.GET_TWITTER_POSTS_BY_AUTHOR, args
        )

    async def search_posts(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        author_username: str | None = None,
        author_id: str | None = None,
        language: str | None = None,
        force_latest: bool | None = None,
        response_type: str | None = None,
    ) -> AsyncPaginatedResult[TwitterPost]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            authorUsername=author_username,
            authorId=author_id,
            language=language,
            forceLatest=force_latest,
            responseType=response_type,
        )
        if response_type == "csv":
            raw = await self._call_tool(_tools.SEARCH_TWITTER_POSTS, args)
            export_op_id = raw.get("operationId") or raw.get("dataDumpExportOperationId")
            csv_raw: dict[str, Any] = {
                "results": [],
                "dataDumpExportOperationId": export_op_id,
            }
            return await self._build_paginated_result(
                csv_raw, TwitterPost, _tools.SEARCH_TWITTER_POSTS, args
            )
        result = await self._call_and_maybe_poll(_tools.SEARCH_TWITTER_POSTS, args)
        return await self._build_paginated_result(
            result, TwitterPost, _tools.SEARCH_TWITTER_POSTS, args
        )

    async def get_retweets(
        self,
        post_id: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
    ) -> AsyncPaginatedResult[TwitterPost]:
        args = self._build_args(
            postId=post_id,
            fields=self._convert_fields(fields),
            startDate=start_date,
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_RETWEETS, args)
        return await self._build_paginated_result(
            result, TwitterPost, _tools.GET_TWITTER_RETWEETS, args
        )

    async def get_quotes(
        self,
        post_id: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[TwitterPost]:
        args = self._build_args(
            postId=post_id,
            fields=self._convert_fields(fields),
            startDate=start_date,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_QUOTES, args)
        return await self._build_paginated_result(
            result, TwitterPost, _tools.GET_TWITTER_QUOTES, args
        )

    async def get_comments(
        self,
        post_id: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[TwitterPost]:
        args = self._build_args(
            postId=post_id,
            fields=self._convert_fields(fields),
            startDate=start_date,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_COMMENTS, args)
        return await self._build_paginated_result(
            result, TwitterPost, _tools.GET_TWITTER_COMMENTS, args
        )

    async def get_post_interacting_users(
        self,
        post_id: str,
        interaction_type: str,
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[TwitterUser]:
        args = self._build_args(
            postId=post_id,
            interactionType=interaction_type,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_POST_INTERACTING_USERS, args)
        return await self._build_paginated_result(
            result, TwitterUser, _tools.GET_TWITTER_POST_INTERACTING_USERS, args
        )

    async def count_posts(
        self,
        phrase: str,
        *,
        start_date: str | None = None,
        end_date: str | None = None,
    ) -> int:
        args = self._build_args(
            phrase=phrase,
            startDate=start_date,
            endDate=end_date,
        )
        result = await self._call_and_maybe_poll(_tools.COUNT_TWEETS, args)
        count = result.get("results", result.get("count", 0))
        if isinstance(count, list):
            if len(count) == 0:
                return 0
            first = count[0]
            if isinstance(first, dict):
                return int(next(iter(first.values())))
            return int(first)
        return int(count)

    async def get_user(
        self,
        identifier: str,
        identifier_type: str = "username",
        *,
        fields: list[str] | None = None,
    ) -> TwitterUser:
        args = self._build_args(
            identifier=identifier,
            identifierType=identifier_type,
            fields=self._convert_fields(fields),
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_USER, args)
        results = result.get("results", [])
        if isinstance(results, list) and len(results) > 0:
            return _parse_item(TwitterUser, results[0])
        return _parse_item(TwitterUser, result)

    async def search_users(
        self,
        name: str,
        *,
        limit: int | None = None,
        fields: list[str] | None = None,
    ) -> list[TwitterUser]:
        args = self._build_args(
            name=name,
            limit=limit,
            fields=self._convert_fields(fields),
        )
        result = await self._call_and_maybe_poll(_tools.SEARCH_TWITTER_USERS, args)
        return _parse_items(TwitterUser, result.get("results", []))

    async def get_user_connections(
        self,
        username: str,
        connection_type: str,
        *,
        fields: list[str] | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[TwitterUser]:
        args = self._build_args(
            username=username,
            connectionType=connection_type,
            fields=self._convert_fields(fields),
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_USER_CONNECTIONS, args)
        return await self._build_paginated_result(
            result, TwitterUser, _tools.GET_TWITTER_USER_CONNECTIONS, args
        )

    async def get_users_by_keywords(
        self,
        query: str,
        *,
        fields: list[str] | None = None,
        start_date: str | None = None,
        end_date: str | None = None,
        language: str | None = None,
        force_latest: bool | None = None,
    ) -> AsyncPaginatedResult[TwitterUser]:
        args = self._build_args(
            query=query,
            fields=self._convert_fields(fields),
            startDate=start_date,
            endDate=end_date,
            language=language,
            forceLatest=force_latest,
        )
        result = await self._call_and_maybe_poll(_tools.GET_TWITTER_USERS_BY_KEYWORDS, args)
        return await self._build_paginated_result(
            result, TwitterUser, _tools.GET_TWITTER_USERS_BY_KEYWORDS, args
        )



from __future__ import annotations

import os
from typing import Any

from xpoz._mcp._transport import SyncTransport
from xpoz._mcp._polling import DEFAULT_TIMEOUT_SECONDS
from xpoz._exceptions import AuthenticationError
from xpoz._config._constants import DEFAULT_SERVER_URL, ENV_API_KEY, ENV_SERVER_URL
from xpoz.namespaces.twitter import TwitterNamespace
from xpoz.namespaces.instagram import InstagramNamespace
from xpoz.namespaces.reddit import RedditNamespace


class XpozClient:
    def __init__(
        self,
        api_key: str | None = None,
        *,
        server_url: str | None = None,
        timeout: float = DEFAULT_TIMEOUT_SECONDS,
    ):
        self._api_key = api_key or os.environ.get(ENV_API_KEY)
        if not self._api_key:
            raise AuthenticationError(
                f"API key required. Get your token at http://xpoz.ai/get-token?utm_source=python_sdk&utm_medium=sdk "
                f"(login → copy token), then pass it as api_key= or set the {ENV_API_KEY} environment variable."
            )

        self._server_url = server_url or os.environ.get(ENV_SERVER_URL) or DEFAULT_SERVER_URL
        self._timeout = timeout
        self._transport = SyncTransport(self._server_url, self._api_key)
        self._transport.connect()

        self.twitter = TwitterNamespace(self._transport.call_tool, self._timeout)
        self.instagram = InstagramNamespace(self._transport.call_tool, self._timeout)
        self.reddit = RedditNamespace(self._transport.call_tool, self._timeout)

    def close(self) -> None:
        self._transport.close()

    def __enter__(self) -> XpozClient:
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()



import pytest

from xpoz import PaginatedResult
from xpoz.types.reddit import (
    RedditPost,
    RedditUser,
    RedditComment,
    RedditSubreddit,
    RedditPostWithComments,
    SubredditWithPosts,
)
from xpoz.types.common import PaginationInfo


@pytest.fixture(scope="module")
def reddit_search_result(client):
    return client.reddit.search_posts(
        "python", fields=["id", "title", "score"]
    )


@pytest.fixture(scope="module")
def reddit_post_id():
    return "1l4da15"


class TestRedditUsers:
    def test_get_user(self, client):
        user = client.reddit.get_user("spez")
        assert isinstance(user, RedditUser)
        if getattr(user, "error", None):
            pytest.skip(f"Server error: {user.error}")
        assert user.username == "spez"

    def test_search_users(self, client):
        users = client.reddit.search_users("spez")
        assert isinstance(users, list)
        assert len(users) > 0
        for u in users:
            assert isinstance(u, RedditUser)

    def test_get_users_by_keywords(self, client):
        result = client.reddit.get_users_by_keywords("programming")
        assert isinstance(result, PaginatedResult)
        assert len(result.data) > 0
        for u in result.data:
            assert isinstance(u, RedditUser)


class TestRedditPosts:
    def test_search_posts(self, reddit_search_result):
        result = reddit_search_result
        assert isinstance(result, PaginatedResult)
        assert isinstance(result.pagination, PaginationInfo)
        assert result.pagination.total_rows > 0
        assert result.pagination.page_number == 1
        assert len(result.data) > 0
        for post in result.data:
            assert isinstance(post, RedditPost)

    def test_search_posts_with_subreddit(self, client):
        result = client.reddit.search_posts(
            "help", subreddit="python", fields=["id", "title", "subreddit_name"]
        )
        assert isinstance(result, PaginatedResult)
        assert len(result.data) > 0

    def test_search_posts_pagination(self, reddit_search_result):
        if not reddit_search_result.has_next_page():
            pytest.skip("Only one page of results")
        page2 = reddit_search_result.next_page()
        assert isinstance(page2, PaginatedResult)
        assert len(page2.data) > 0

    def test_get_post_with_comments(self, client, reddit_post_id):
        result = client.reddit.get_post_with_comments(reddit_post_id)
        assert isinstance(result, RedditPostWithComments)
        assert isinstance(result.post, RedditPost)
        assert isinstance(result.comments, list)

    def test_search_comments(self, client):
        result = client.reddit.search_comments("python")
        assert isinstance(result, PaginatedResult)
        assert len(result.data) > 0
        for c in result.data:
            assert isinstance(c, RedditComment)


class TestRedditSubreddits:
    def test_search_subreddits(self, client):
        subreddits = client.reddit.search_subreddits("python")
        assert isinstance(subreddits, list)
        if len(subreddits) == 0:
            pytest.skip("Server returned empty results (transient)")
        for s in subreddits:
            assert isinstance(s, RedditSubreddit)

    def test_get_subreddit_with_posts(self, client):
        result = client.reddit.get_subreddit_with_posts("python")
        assert isinstance(result, SubredditWithPosts)
        assert isinstance(result.subreddit, RedditSubreddit)
        assert result.subreddit.display_name is not None
        assert isinstance(result.posts, list)

    def test_get_subreddits_by_keywords(self, client):
        result = client.reddit.get_subreddits_by_keywords("programming")
        assert isinstance(result, PaginatedResult)
        assert len(result.data) > 0
        for s in result.data:
            assert isinstance(s, RedditSubreddit)



from xpoz._client import XpozClient
from xpoz._async_client import AsyncXpozClient
from xpoz._exceptions import (
    XpozError,
    AuthenticationError,
    XpozConnectionError,
    OperationTimeoutError,
    OperationFailedError,
    OperationCancelledError,
    NotFoundError,
    ValidationError,
)
from xpoz._pagination import PaginatedResult, AsyncPaginatedResult
from xpoz._version import __version__

__all__ = [
    "XpozClient",
    "AsyncXpozClient",
    "XpozError",
    "AuthenticationError",
    "XpozConnectionError",
    "OperationTimeoutError",
    "OperationFailedError",
    "OperationCancelledError",
    "NotFoundError",
    "ValidationError",
    "PaginatedResult",
    "AsyncPaginatedResult",
    "__version__",
]



from pydantic import BaseModel


class PaginationInfo(BaseModel, extra="allow"):
    table_name: str | None = None
    total_rows: int = 0
    total_pages: int = 0
    page_number: int = 1
    page_size: int | None = None
    results_count: int = 0



from xpoz.namespaces.twitter import TwitterNamespace, AsyncTwitterNamespace
from xpoz.namespaces.instagram import InstagramNamespace, AsyncInstagramNamespace
from xpoz.namespaces.reddit import RedditNamespace, AsyncRedditNamespace

__all__ = [
    "TwitterNamespace",
    "AsyncTwitterNamespace",
    "InstagramNamespace",
    "AsyncInstagramNamespace",
    "RedditNamespace",
    "AsyncRedditNamespace",
]



from xpoz.types.common import PaginationInfo
from xpoz.types.twitter import TwitterPost, TwitterUser
from xpoz.types.instagram import InstagramPost, InstagramUser, InstagramComment
from xpoz.types.reddit import (
    RedditPost,
    RedditUser,
    RedditComment,
    RedditSubreddit,
    RedditPostWithComments,
    SubredditWithPosts,
)

__all__ = [
    "PaginationInfo",
    "TwitterPost",
    "TwitterUser",
    "InstagramPost",
    "InstagramUser",
    "InstagramComment",
    "RedditPost",
    "RedditUser",
    "RedditComment",
    "RedditSubreddit",
    "RedditPostWithComments",
    "SubredditWithPosts",
]



from __future__ import annotations

from typing import Any

import anyio
import httpx
from anyio.from_thread import BlockingPortal, start_blocking_portal
from mcp import ClientSession
from mcp.client.streamable_http import streamable_http_client

from xpoz._transform._response_parser import parse_response_text
from xpoz._version import __version__

_USER_AGENT = f"xpoz-python-sdk/{__version__}"


def _parse_tool_result(tool_name: str, result: Any) -> dict[str, Any]:
    if result.isError:
        error_text = ""
        for block in result.content:
            if hasattr(block, "text"):
                error_text += block.text
        raise RuntimeError(f"MCP tool error ({tool_name}): {error_text}")

    combined_text = ""
    for block in result.content:
        if hasattr(block, "text"):
            combined_text += block.text

    return parse_response_text(combined_text)


class McpTransport:
    def __init__(self, server_url: str, api_key: str | None = None):
        self._server_url = server_url
        self._api_key = api_key
        self._session: ClientSession | None = None
        self._context_stack: list[Any] = []

    async def connect(self) -> None:
        headers: dict[str, str] = {"User-Agent": _USER_AGENT}
        if self._api_key:
            headers["Authorization"] = f"Bearer {self._api_key}"

        http_client = httpx.AsyncClient(
            headers=headers,
            timeout=httpx.Timeout(30, read=None),
        )
        ctx = streamable_http_client(self._server_url, http_client=http_client)
        streams = await ctx.__aenter__()
        self._context_stack.append(ctx)

        read_stream, write_stream, _ = streams
        session_ctx = ClientSession(read_stream, write_stream)
        self._session = await session_ctx.__aenter__()
        self._context_stack.append(session_ctx)

        await self._session.initialize()

    async def close(self) -> None:
        for ctx in reversed(self._context_stack):
            try:
                await ctx.__aexit__(None, None, None)
            except Exception:
                pass
        self._context_stack.clear()
        self._session = None

    async def call_tool(self, tool_name: str, arguments: dict[str, Any]) -> dict[str, Any]:
        if self._session is None:
            raise RuntimeError("Transport not connected. Call connect() first.")

        result = await self._session.call_tool(tool_name, arguments)
        return _parse_tool_result(tool_name, result)


class SyncTransport:
    def __init__(self, server_url: str, api_key: str | None = None):
        self._server_url = server_url
        self._api_key = api_key
        self._session: ClientSession | None = None
        self._portal_cm: Any = None
        self._portal: BlockingPortal | None = None
        self._shutdown_event: anyio.Event | None = None
        self._connect_error: BaseException | None = None

    def connect(self) -> None:
        self._portal_cm = start_blocking_portal()
        self._portal = self._portal_cm.__enter__()

        ready_event = self._portal.call(anyio.Event)
        self._shutdown_event = self._portal.call(anyio.Event)
        self._portal.start_task_soon(self._lifecycle, ready_event, self._shutdown_event)
        self._portal.call(ready_event.wait)

        if self._connect_error is not None:
            raise self._connect_error

    async def _lifecycle(
        self,
        ready: anyio.Event,
        shutdown: anyio.Event,
    ) -> None:
        try:
            headers: dict[str, str] = {"User-Agent": _USER_AGENT}
            if self._api_key:
                headers["Authorization"] = f"Bearer {self._api_key}"

            http_client = httpx.AsyncClient(
                headers=headers,
                timeout=httpx.Timeout(30, read=None),
            )
            async with streamable_http_client(self._server_url, http_client=http_client) as streams:
                read_stream, write_stream, _ = streams
                async with ClientSession(read_stream, write_stream) as session:
                    await session.initialize()
                    self._session = session
                    ready.set()
                    await shutdown.wait()
        except BaseException as exc:
            self._connect_error = exc
            ready.set()
            raise
        finally:
            self._session = None

    def close(self) -> None:
        if self._portal and self._shutdown_event:
            try:
                self._portal.call(self._shutdown_event.set)
            except Exception:
                pass
        if self._portal_cm:
            try:
                self._portal_cm.__exit__(None, None, None)
            except Exception:
                pass
            self._portal_cm = None
            self._portal = None
            self._session = None

    def call_tool(self, tool_name: str, arguments: dict[str, Any]) -> dict[str, Any]:
        if self._portal is None or self._session is None:
            raise RuntimeError("Transport not connected. Call connect() first.")

        async def _call() -> dict[str, Any]:
            assert self._session is not None
            result = await self._session.call_tool(tool_name, arguments)
            return _parse_tool_result(tool_name, result)

        return self._portal.call(_call)
